module counter ! counter for jobs
  integer :: jobs_done=0
  integer :: m,n
  real*8 :: dre,dim
end module counter

program slaves
  ! all-slave model programmed in OpenMP
  ! getting the job is done sequentially (critical)
  ! executing it is done sin parallel
  use omp_lib
  use counter
  implicit none
  integer :: my_job,res,count=0
  read(*,*)m,n
  dre=0.d0
  dim=0.d0
  if (m.gt.0) dre=2.d0/m
  if (n.gt.0) dim=1.d0/n
  !$omp parallel private(my_job)
  call get_job(my_job)     ! initial job for each thread
  do while (my_job.ne.0) ! [do job, get job] as long as there are any
     call job(my_job,omp_get_thread_num(),res)
     count=count+res
     call get_job(my_job)
  end do
  !$omp end parallel
  write(*,*)' Counter=',count
end program slaves

subroutine get_job(my_job)
! job assignment
    use counter
    integer :: my_job
! avoiding race condition through critical
!$omp critical 
    my_job=0
    if (jobs_done.lt.n+1) then
       jobs_done=jobs_done+1
       my_job=jobs_done
    end if
!$omp end critical
    return
end subroutine get_job

subroutine job(number,thread,res)
  ! A "line" of the Mandelbrot set
  use counter
  implicit none
  integer, intent(in) :: number,thread
  integer, intent(out) :: res
  integer :: i,set
  real*8 :: re,lim
  lim=(number-1)*dim
  res=0
  do i=0,m
     re=i*dre-1.5d0
     res=res+set(re,lim)
     ! write(*,*)i,number-1,re,lim,res
  end do
!  write(*,101)number,thread,lim
101 format(' Job number ',i2,' done by thread number ',i2,' imaginary part =',f14.8)
  return
end subroutine job

integer function set(re,im)
  implicit none
  real*8, intent(in) :: re, im
  real*8 :: testr, testi, testrn, testin, test
  integer :: i
  integer, parameter :: limit=1024
  set=1
  testr=0.d0
  testi=0.d0
  testrn=0.d0
  testin=0.d0
  do i=1,limit
     testrn=re+testr*testr-testi*testi
     testin=im+2*testi*testr
     testr=testrn
     testi=testin
     test=testr*testr+testi*testi
     if(test.gt.4.d0) then
        set=0
        exit
     end if
  end do
  return
end function set
