!-------------------------------------------------------------------
! Matrix multiplication code for the course .Practical OpenMP.
! It performs parallelized multiplication of matrix A with matrix B.
! HPCVL, Queen.s University, May, 2006
!-------------------------------------------------------------------
module matrices           ! Info about matrices, I/O, (de)allocation
  integer:: msz(3)
  real*8, allocatable, dimension(:,:) :: ma,mb,mc
contains 
  
  subroutine mtxalloc     ! Allocationg memory for A,B,C
    allocate(ma(msz(3),msz(2))) 
    allocate(mb(msz(2),msz(1)))
    allocate(mc(msz(3),msz(1)))
  end subroutine mtxalloc
  
  subroutine mtxdealloc   ! Deallocation of A,B,C
    deallocate(ma,mb,mc)   
  end subroutine mtxdealloc
  
  subroutine mtxin        ! Reading matrices in
    open(1,file='in.dat')
    read(1,*) msz(3), msz(2), msz(1) ! Reading dimensions
    call mtxalloc         ! Memory allocation
    read(1,*)((ma(i,j),j=1,msz(2)),i=1,msz(3)) ! Reading A and B
    read(1,*)((mb(i,j),j=1,msz(1)),i=1,msz(2))
    mc=0.0d0              ! Initializing result matrix C
    close(1)
    return
  end subroutine mtxin
  
  subroutine mtxout       ! Output of result (serial)
    open(1,file='out.dat') 
    do j=1, msz(3)
       write(1,*)(mc(j,i),i=1,msz(1))
    end do
    close(1)
    call mtxdealloc       ! Deallocation
  end subroutine mtxout

end module matrices
!-------------------------------------------------------------------
program matmul            ! Calling routine (wrapper)
  call multiplication     ! Performing multiplication
  stop
end program matmul
!-------------------------------------------------------------------
subroutine multiplication ! Main work
  use matrices            ! Info about matrices
  implicit none
  integer :: i,j,k

  call mtxin              ! Reading and allocating matrices

!$omp parallel do private(i,j,k) shared(ma,mb,mc,msz)
  do i=1,msz(3)
     do j=1,msz(1)
        do k=1,msz(2)
           mc(i,j)=mc(i,j)+ma(i,k)*mb(k,j)
        end do
     end do
  end do
!$omp end parallel do

  call mtxout             ! Output of the result C, deallocation

  return
end subroutine multiplication
!-------------------------------------------------------------------
