module work
  integer, dimension(20) :: sleeptime
contains
  subroutine load
    integer :: i
    real :: r
    do i=1,20
       call random_number(r)
       sleeptime(i)=1+int(4*r)
    end do
  end subroutine load
end module work

program slaves
  ! all-slave model programmed in OpenMP
  ! getting the job is done sequentially (critical)
  ! executing it is done sin parallel
  use work
  implicit none
  integer :: my_job,get_job,omp_get_thread_num
  call load
  !$omp parallel private(my_job)
  my_job = get_job()     ! initial job for each thread
  do while (my_job.ne.0) ! [do job, get job] as long as there are any
     call job(my_job,omp_get_thread_num())
     my_job=get_job()
  end do
  !$omp end parallel
end program slaves

module counter ! counter for jobs
  integer :: jobs_done=0
end module counter

integer function get_job()
  ! job assignment
  use counter
  integer, parameter :: num_jobs=20
  ! avoiding race condition through critical
  !$omp critical
  get_job=0
  if (jobs_done.lt.num_jobs) then
     jobs_done=jobs_done+1
     get_job=jobs_done
  end if
  !$omp end critical
  return
end function get_job

subroutine job(number,thread)
  ! do senseless busy-work, report back
  use work
  implicit none
  integer :: number,thread,i,st
  st=sleeptime(number) 
  call sleep(st)
  write(*,101)number,thread,st
101 format(' Job number ',i2,' done by thread number ',i2,' result =',i3)
  return
end subroutine job
